\ ASN1DER.f  ASN.1 Abstract Syntax Notation version 1 Distinguished Encoding Rules 
\ 
\ Download, install and run SwiftForth from www.forth.com
\ cd to the directory containg ASN1DER.f and its support files from the ASN1DER.7z file
\ type "include ASN1DER.f"    ( without the quotes ). Press Enter.
\ 
\ Parse an X.509 Certificate that uses ASN.1 Abstract Syntax Notation version 1 and DER Distinguished Encoding Rules   
\ DER = BER Basic Encoding Rules with special extra rules : 
\ the length format must be the smallest possible, cannot be an indefinite length, attributes will always be in the same order, etc.
\ For simple string types and implicitly tagged types derived from simple string types, the primitive, definite-length method must be employed.
\ For structured types, implicitly tagged types derived from structured types, and explicitly tagged types derived from anything, the constructed, 
\ definite-length method must be employed. From : http://luca.ntop.org/Teaching/Appunti/asn1.html
\ ( A Layman's Guide to a Subset of ASN.1, BER, and DER, An RSA Laboratories Technical Note, Burton S. Kaliski Jr., Revised November 1, 1993 )
\ A certificate follows X.509 : http://en.wikipedia.org/wiki/X.509  http://tools.ietf.org/html/rfc3280
\ Encoded using ASN.1 and DER : http://en.wikipedia.org/wiki/X.690  ITU-T X.690, ISO/IEC 8825-1  http://www.itu.int/ITU-T/studygroups/com17/languages/X.690-0207.pdf
\ signature algorithms        : https://tools.ietf.org/html/rfc5758
\ Object IDentifiers          : http://oid-info.com/index.htm   https://docs.oracle.com/javame/8.0/api/gcf/api/javax/microedition/pki/Certificate.html

\ OID values encode the first 2 values in the first octet so the decimal dot form 2.5.4.11 --> 55 04 0B
\ To find the meaning of an OID, decode the OID string (see below) and go to http://oid-info.com/

\ From : http://luca.ntop.org/Teaching/Appunti/asn1.html :
\ The first octet has value 40 * value1 + value2. 
\ (This is unambiguous, since value1 is limited to values 0, 1, and 2; value2 is limited to the range 0 to 39 when value1 is 0 or 1; and, according to X.208, n is always at least 2.)
\ The following octets, if any, encode value3, ..., valuen. 
\ Each value is encoded base 128, most significant digit first, with as few digits as possible, 
\ and the most significant bit of each octet except the last in the value's encoding set to "1."
\ Example: The first octet of the BER encoding of RSA Data Security, Inc.'s object identifier is
\ 40 * 1 + 2 = 42 = 2a16. 
\ The encoding of 840 = 6 * 128 + 0x48 is 86 48 and the encoding of 113549 = 6 * 1282 + 0x77 * 128 + 0x0d is 86 f7 0d. 
\ This leads to the following BER encoding:
\ 06 06 2a 86 48 86 f7 0d

\ Note : we do not need to decode the Certificate OIDs - we just compare them to the pre-encoded strings below - above if for information only.
\ Note : the first octet of each string below is its length.

ONLY FORTH ALSO DEFINITIONS DECIMAL

empty

: HostVersion ( -- a n )   s" ASN1DER 2015 Dec 17" ; 

\ The strings below are used to by MSwindow.f 
: Acronym  HostVersion  drop 7 ;   \ must not have any spaces in the acronym 
create AboutHeader  ," About ASN1DER" 
create AboutLine0   ," ASN1 Parser"
create AboutLine1   HostVersion STRING,
create AboutLine2   ,"  2015 Inventio Software"
create AboutLine3   ," www.inventio.co.uk"  
create HelpFilename ," ASN1DER_help.pdf"

: DefaultSaveFilename   s" ASN1DER_saveFile" ;

MAX_PATH constant |CapturedFilename| 
|CapturedFilename| Buffer: CapturedFilename
: CapturedFilenameInit   s" ASN1DER.txt" CapturedFilename place ;
CapturedFilenameInit

MAX_PATH constant |SaveFilename| 
|SaveFilename| Buffer: SaveFilename
DefaultSaveFilename SaveFilename place 

variable indent         \ the current indentation level
variable StructLevel    \ where we are in the elements in the structure

3 constant INDENT_SPACES
\ 4 constant INDENT_SPACES

\ does a cr ( new line ) then indents by a number of spaces
: cr+   cr  indent @ StructLevel @ if  1  else  0  then  + INDENT_SPACES * spaces ;

s" Support.f" FILE-STATUS swap drop [IF]    \ look for the local copy of the support file, if not there, load from the Common folder
cr .( Using ..\Common\Support.f )

include ..\Common\MD5.f
include ..\Common\NumName.f
include ..\Common\Support.f
        
include ..\Common\SHA-256.f
include ..\Common\LogToFile.f
include ..\Common\CRC32.f
include ..\Common\Walk.f
include ..\Common\ChooseFile.f
include ..\Common\MD5.f

[ELSE]

cr .( Using .\Support.f )

include MD5.f
include NumName.f
include Support.f

include SHA-256.f
include LogToFile.f
include CRC32.f
include Walk.f
include ChooseFile.f

[THEN]

: InitSaveFilename ( -- )    Acronym SaveFilename place   s" _SaveFile" SaveFilename append 
   AppendDate&TimeToSaveFilename  s" .txt" SaveFilename append
;
InitSaveFilename

\ *****************************************************************************
\ End of common code 
\ *****************************************************************************

include ASN1DERerrors.f

\ include gcm.f  \ Galois field multiplication and Counter Mode ToDo!!!

\ *****************************************************************************
\ *****************************************************************************

#12 2** constant |PayloadData| 
|PayloadData| Qbuffer: PayloadData[]
Qvariable WritePayloadDataOffset
Qvariable ReadPayloadDataOffset

: >PayloadData ( n -- a )   |PayloadData| 1- and PayloadData[] + ;  

: PayloadDataPut8 ( c -- ) 
   WritePayloadDataOffset @ >PayloadData c! 
   1 WritePayloadDataOffset +! 
;

: PayloadDataGet8 ( -- c )
   ReadPayloadDataOffset @ WritePayloadDataOffset @ < if
      ReadPayloadDataOffset @ >PayloadData c@
      1 ReadPayloadDataOffset +! 
   else
      #10002 throw  \ run out of data
   then
;

: ttShowPayloadData ( -- )
   0 ReadPayloadDataOffset !
   WritePayloadDataOffset @ 0 ?do  
      i $0F and 0= if  cr  3 spaces  then
      PayloadDataGet8 2.hex space  
   loop 
;

: PayloadDataInit ( -- )   0 WritePayloadDataOffset !  0 ReadPayloadDataOffset ! ;

PayloadDataInit

: ttpld ( -- )
   PayloadDataInit  5 0 do  i [char] a + PayloadDataPut8 loop  ttShowPayloadData 
;

\ Maximum number of intermediate CAs in a verification chain.
\ That is, maximum length of the chain, excluding the end-entity certificate
\ and the trusted root certificate.
\ 
\ Set this to a low value to prevent an adversary from making you waste
\ resources verifying an overlong certificate chain.
#define POLARSSL_X509_MAX_INTERMEDIATE_CA   8


\ X509 Verify codes
    $01 constant BADCERT_EXPIRED          \ The certificate validity has expired
    $02 constant BADCERT_REVOKED          \ The certificate has been revoked (is on a CRL)
    $04 constant BADCERT_CN_MISMATCH      \ The certificate Common Name (CN) does not match with the expected CN
    $08 constant BADCERT_NOT_TRUSTED      \ The certificate is not correctly signed by the trusted CA
    $10 constant BADCRL_NOT_TRUSTED       \ CRL is not correctly signed by the trusted CA
    $20 constant BADCRL_EXPIRED           \ CRL is expired
    $40 constant BADCERT_MISSING          \ Certificate was missing
    $80 constant BADCERT_SKIP_VERIFY      \ Certificate verification was skipped
  $0100 constant BADCERT_OTHER            \ Other reason (can be used by verify callback)
  $0200 constant BADCERT_FUTURE           \ The certificate validity starts in the future
  $0400 constant BADCRL_FUTURE            \ The CRL is from the future
         
\ X.509 v3 Key Usage Extension flags

$80 constant KU_DIGITAL_SIGNATURE              \ bit 0
$40 constant KU_NON_REPUDIATION                \ bit 1
$20 constant KU_KEY_ENCIPHERMENT               \ bit 2
$10 constant KU_DATA_ENCIPHERMENT              \ bit 3
$08 constant KU_KEY_AGREEMENT                  \ bit 4
$04 constant KU_KEY_CERT_SIGN                  \ bit 5
$02 constant KU_CRL_SIGN                       \ bit 6

\ Netscape certificate types (http://www.mozilla.org/projects/security/pki/nss/tech-notes/tn3.html)
$80 constant NS_CERT_TYPE_SSL_CLIENT           \ bit 0
$40 constant NS_CERT_TYPE_SSL_SERVER           \ bit 1
$20 constant NS_CERT_TYPE_EMAIL                \ bit 2
$10 constant NS_CERT_TYPE_OBJECT_SIGNING       \ bit 3
$08 constant NS_CERT_TYPE_RESERVED             \ bit 4
$04 constant NS_CERT_TYPE_SSL_CA               \ bit 5
$02 constant NS_CERT_TYPE_EMAIL_CA             \ bit 6
$01 constant NS_CERT_TYPE_OBJECT_SIGNING_CA    \ bit 7


\ X.509 extension types
\ Comments refer to the status for using certificates. Status can be
\ different for writing certificates or reading CRLs or CSRs.
    $01 constant EXT_AUTHORITY_KEY_IDENTIFIER   
    $02 constant EXT_SUBJECT_KEY_IDENTIFIER     
    $04 constant EXT_KEY_USAGE                   \ Parsed but not used
    $08 constant EXT_CERTIFICATE_POLICIES       
    $10 constant EXT_POLICY_MAPPINGS            
    $20 constant EXT_SUBJECT_ALT_NAME            \ Supported (DNS)
    $40 constant EXT_ISSUER_ALT_NAME            
    $80 constant EXT_SUBJECT_DIRECTORY_ATTRS    
  $0100 constant EXT_BASIC_CONSTRAINTS           \ Supported
  $0200 constant EXT_NAME_CONSTRAINTS           
  $0400 constant EXT_POLICY_CONSTRAINTS         
  $0800 constant EXT_EXTENDED_KEY_USAGE          \ Parsed but not used
  $1000 constant EXT_CRL_DISTRIBUTION_POINTS    
  $2000 constant EXT_INIHIBIT_ANYPOLICY         
  $4000 constant EXT_FRESHEST_CRL               
        
 $10000 constant EXT_NS_CERT_TYPE                \ Parsed (and then ?)


\ Storage format identifiers - PEM and DER
   1 constant X509_FORMAT_DER              
   2 constant X509_FORMAT_PEM              
      
$100 constant X509_MAX_DN_NAME_SIZE             \ Maximum value size of a DN entry


\ http://en.wikipedia.org/wiki/X.690 defines the ASN.1 BER encoding, DER adds some exxtra rules such as : the length encoding must be as small as possible

\ DER encoded ANS.1 type tags :
$01 constant ASN1_BOOLEAN                 
$02 constant ASN1_INTEGER                 
$03 constant ASN1_BIT_STRING              
$04 constant ASN1_OCTET_STRING            
$05 constant ASN1_NULL                    
$06 constant ASN1_OID                     
$0C constant ASN1_UTF8_STRING             
$10 constant ASN1_SEQUENCE                
$11 constant ASN1_SET                     
$13 constant ASN1_PRINTABLE_STRING        
$14 constant ASN1_T61_STRING              
$16 constant ASN1_IA5_STRING              
$17 constant ASN1_UTC_TIME                
$18 constant ASN1_GENERALIZED_TIME        
$1C constant ASN1_UNIVERSAL_STRING        
$1E constant ASN1_BMP_STRING   

\ Bit 5 true specifies that the Tag is constructed, else primitive
$00 constant ASN1_PRIMITIVE   
$20 constant ASN1_CONSTRUCTED
\ bits 6 and 7 specify that the Class of the Tag 
$80 constant ASN1_CONTEXT_SPECIFIC        

: NotAllowed ( -- )   RED. ." NotAllowed!! "  RESTORE. ;

variable ConstructedPacket
variable CurrentTag
variable CurrentTagPtr            \ pointer to the start of the current Tag's Value field
variable CurrentTagLength         \ The current Tag's value's length

\ Type Length Value structure 
: ShowTag ( c -- )
   MAGENTA.  dup 2.hex  space  RESTORE.
   >r
   r@ 6 rshift $03 and case   \ Class of Tag
      $00 of  ." Universal       "  endof  \ The type is native to ASN.1
      $01 of  ." Application     "  endof  \ The type is only valid for one specific application
      $02 of  ." Context-specific"  endof  \ The meaning of this type depends on the context (such as within a sequence, set or choice)
      $03 of  ." Private         "  endof  \ Defined in private specification
   endcase                                                                        \  Constructed  
\   r@ ASN1_CONSTRUCTED and dup ConstructedPacket !  if  ."  Constructed "  else  ."  Primitive   "  then
   r@ ASN1_CONSTRUCTED and dup ConstructedPacket !  if  ."  C "  else  ."  P "  then

   r@ $3F and case
      $00 of  ." EOC (End-of-Content)"  endof \   P   
      $01 of  ." BOOLEAN	          "  endof \   P   
      $02 of  ." INTEGER	          "  endof \   P   
      $03 of  ." BIT STRING	       "  endof \   P/C	
      $04 of  ." OCTET STRING	       "  endof \   P/C	
      $05 of  ." NULL	             "  endof \   P   
      $06 of  MAGENTA. ." OBJECT IDENTIFIER	 "  RESTORE.  endof \   P   
      $07 of  ." Object Descriptor	 "  endof \   P/C	
      $08 of  ." EXTERNAL	          "  endof \   C   
      $09 of  ." REAL (float)	       "  endof \   P   
      $0A of  ." ENUMERATED	       "  endof \   P   
      $0B of  NotAllowed ." EMBEDDED PDV	       "  endof \   C   
      $0C of  ." UTF8String	       "  endof \   P/C	
      $0D of  ." RELATIVE-OID	       "  endof \   P   
      $0E of  ." (reserved)	       "  endof \       
      $0F of  ." (reserved)	       "  endof \       
      $10 of  NotAllowed ." SEQUENCE 	          "  endof \   C   
      $11 of  NotAllowed ." SET                 "  endof \   C   
      $12 of  ." NumericString	    "  endof \   P/C	
      $13 of  ." PrintableString	    "  endof \   P/C	
      $14 of  ." T61String	          "  endof \   P/C	
      $15 of  ." VideotexString	    "  endof \   P/C	
      $16 of  ." IA5String	          "  endof \   P/C	
      $17 of  ." UTCTime	          "  endof \   P/C	
      $18 of  ." GeneralizedTime	    "  endof \   P/C	
      $19 of  ." GraphicString	    "  endof \   P/C	
      $1A of  ." VisibleString	    "  endof \   P/C	
      $1B of  ." GeneralString	    "  endof \   P/C	
      $1C of  ." UniversalString	    "  endof \   P/C	
      $1D of  ." CHARACTER STRING	 "  endof \   P/C	
      $1E of  ." BMPString	          "  endof \   P/C	
      $1F of  ." (use long-form)	 	 "  endof \       

      $20 of  ( NotAllowed ) ." EOC (End-of-Content)"  endof \   P    \ Techincally " NotAllowed " , but seems to be allowed just fine...
      $21 of  NotAllowed ." BOOLEAN	          "  endof \   P    
      $22 of  NotAllowed ." INTEGER	          "  endof \   P    
      $23 of  ." BIT STRING	       "  endof \   P/C	 
      $24 of  ." OCTET STRING	       "  endof \   P/C	 
      $25 of  NotAllowed ." NULL	             "  endof \   P    
      $26 of  NotAllowed ." OBJECT IDENTIFIER	 "  endof \   P    
      $27 of  ." Object Descriptor	 "  endof \   P/C	 
      $28 of  ." EXTERNAL	          "  endof \   C    
      $29 of  NotAllowed ." REAL (float)	       "  endof \   P    
      $2A of  NotAllowed ." ENUMERATED	       "  endof \   P    
      $2B of  ." EMBEDDED PDV	       "  endof \   C    
      $2C of  ." UTF8String	       "  endof \   P/C	 
      $2D of  NotAllowed ." RELATIVE-OID	       "  endof \   P    
      $2E of  ." (reserved)	       "  endof \        
      $2F of  ." (reserved)	       "  endof \        
      $30 of  ." SEQUENCE 	          "  endof \   C     
      $31 of  ." SET                 "  endof \   C    
      $32 of  ." NumericString	    "  endof \   P/C	 
      $33 of  ." PrintableString	    "  endof \   P/C	 
      $34 of  ." T61String	          "  endof \   P/C	  
      $35 of  ." VideotexString	    "  endof \   P/C	 
      $36 of  ." IA5String	          "  endof \   P/C	 
      $37 of  ." UTCTime	          "  endof \   P/C	 
      $38 of  ." GeneralizedTime	    "  endof \   P/C	  
      $39 of  ." GraphicString	    "  endof \   P/C	 
      $3A of  ." VisibleString	    "  endof \   P/C	 
      $3B of  ." GeneralString	    "  endof \   P/C	 
      $3C of  ." UniversalString	    "  endof \   P/C	   
      $3D of  ." CHARACTER STRING	 "  endof \   P/C	 
      $3E of  ." BMPString	          "  endof \   P/C	 
      $3F of  ." (use long-form)	 	 "  endof \        
   endcase                                                                    
   r> drop                                                                    
;                                      

: ttstag ( -- )
   $45 0 do  cr  i dup 4 .hex  space  BLUE.  ShowTag  RESTORE.  loop 
   cr ." Done :-)"
;

\ *****************************************************************************
\ *****************************************************************************

\ The special ASN.1 variable-length encoded length
: ASN.1GetLength ( -- u ) 
   PayloadDataGet8 dup $80 and if     
      case
      $80 of  -1  endof    \ indefinite length, terminated by and EOC with 0 length ( two 0's )
      $81 of  PayloadDataGet8  endof
      $82 of  PayloadDataGet8 8 lshift PayloadDataGet8 or  endof
      $83 of  PayloadDataGet8 8 lshift PayloadDataGet8 or  8 lshift PayloadDataGet8 or  endof
         cr  RED.  ." Illegal AXDR String length value = "  2 .hex   ."   ReadPayloadDataOffset = "  ReadPayloadDataOffset @ 4 .hex  RESTORE. 
      endcase
   then
; 

\ *****************************************************************************
\ *****************************************************************************



variable StartOfSequence
\ variable EndOfSequence

: xxxShowTLVrecurse ( -- ) 
   cr+  PayloadDataGet8 ShowTag  60 ToColumn ."  Length = " ASN.1GetLength dup BLUE. 4.hex  RESTORE. 2 spaces 
   ConstructedPacket @ if  0 Indent +!  recurse  else  0 Indent +!  then 
\   1 Indent +! 
   0 ?do  i $0F and 0= i 0= not and if  cr+  then PayloadDataGet8 2.hex  space  loop
\   -1 Indent +! 
   recurse
; 

\ *****************************************************************************
\ Sequence end address stack
\ *****************************************************************************

4 2** constant |SequenceEndStack|  \ must be a power of 2
|SequenceEndStack| cells Buffer: SequenceEndStack[]  
variable SequenceEndStackPtr

: >SequenceEndStackPtr ( -- a )   SequenceEndStack[]  SequenceEndStackPtr @  |SequenceEndStack| 1-  and cells + ; 

: SequenceEndPush ( u -- )   1 SequenceEndStackPtr +!  >SequenceEndStackPtr ! ;
: SequenceEndTOS  ( u -- )   >SequenceEndStackPtr @ ;
: SequenceEndPop  ( u -- )   >SequenceEndStackPtr @  -1 SequenceEndStackPtr +! ;

: InitSequenceEndStack ( -- )
   0 SequenceEndStackPtr ! SequenceEndStack[] |SequenceEndStack| cells erase   $7FFFFFFF SequenceEndPush 
;
InitSequenceEndStack


: ShowSequenceStack ( -- )
   SequenceEndStackPtr @ >r
   0 SequenceEndStackPtr !
   cr  ." <--- bottom                                                     SequenceEndStack                                                       top ---> "
   cr  |SequenceEndStack| 0 do  |SequenceEndStack| 1- i - 8 .hex  space  loop    
\   cr  |SequenceEndStack| 0 do  SequenceEndStack[] i cells + @ 8 .hex  space loop    
   cr  BLUE. |SequenceEndStack| 0 do  SequenceEndPop 8 .hex  space loop  RESTORE. 
   r> SequenceEndStackPtr ! 
;

{ 
: ShowSequenceStackVertical ( -- )
   |SequenceEndStack| 0 do  cr i 2 .hex  2 spaces  SequenceEndPop 3 .hex  loop    
;
}

\ *****************************************************************************
\ Count the number of each Tag type in the certificate
\ *****************************************************************************

$40 constant |TagBins|
|TagBins| cells Buffer: TagBins[]

: TagBinsInit ( -- )  TagBins[] |TagBins| cells erase ;

: >TagBin ( c -- a )   |TagBins| 1- and cells TagBins[] + ;

: TagBin+ ( c -- )   1 swap >TagBin +! ; 

: TagBin@ ( c -- u )   >TagBin @ ; 

\ *****************************************************************************
\ The Object Identifiers ( OIDs ) used in an X.509 Certificate
\ *****************************************************************************
{ 
Printable Representation for X.509 Distinguished Names

For a X.509 certificate the value returned is the printable version of the distinguished name (DN) from the certificate.

An X.509 distinguished name of is set of attributes, each attribute is a sequence of an object ID and a value. For string comparison purposes, the following rules define a strict printable representation.

There is no added white space around separators.
The attributes are in the same order as in the certificate; attributes are not reordered.
If an object ID is in the table below, the label from the table will be substituted for the object ID, else the ID is formatted as a string using the binary printable representation above.
Each object ID or label and value within an attribute will be separated by a "=" (Unicode U+003D), even if the value is empty.
If value is not a string, then it is formatted as a string using the binary printable representation above.
Attributes will be separated by a ";" (Unicode U+003B)

Labels for X.500 Distinguished Name Attributes
Object ID                  Binary                     Label
==============================================================
id-at-commonName           55 04 03                   CN
id-at-surname              55 04 04                   SN
id-at-countryName          55 04 06                   C
id-at-localityName         55 04 07                   L
id-at-stateOrProvinceName	55 04 08                   ST
id-at-streetAddress	      55 04 09                   STREET
id-at-organizationName	   55 04 0A                   O
id-at-organizationUnitName	55 04 0B                   OU
emailAddress	            2A 86 48 86 F7 0D 01 09 01	EmailAddress

Example of a printable distinguished name:

C=US;O=Any Company, Inc.;CN=www.anycompany.com
} 

\ From https://www.ietf.org/rfc/rfc5480.txt :
\ id-ecPublicKey indicates that the algorithms that can be used
\    with the subject public key are unrestricted.  The key is only
\    restricted by the values indicated in the key usage certificate
\    extension (see Section 3).  id-ecPublicKey MUST be supported.
\    See Section 2.1.1.  This value is also included in certificates
\    when a public key is used with ECDSA.
\ 
\      id-ecPublicKey OBJECT IDENTIFIER ::= {
\      iso(1) member-body(2) us(840) ansi-X9-62(10045) keyType(2) 1 }
\ 
\     secp256r1/prime256v1(7) OBJECT IDENTIFIER ::= {
\       iso(1) member-body(2) us(840) ansi-X9-62(10045) curves(3)
\       prime(1) 7 }


\ All these are preceded by 30 0A 06 08 - SEQUENCE of length $0A , OBJECT IDENTIFIER of length $08
\ create ecdsa-with-SHA224[] $09 nc, 08  2A 86 48 CE 3D 04 03 01
create ecdsa-with-SHA256[] $09 nc, 08  2A 86 48 CE 3D 04 03 02
\ create ecdsa-with-SHA384[] $09 nc, 08  2A 86 48 CE 3D 04 03 03
\ create ecdsa-with-SHA512[] $09 nc, 08  2A 86 48 CE 3D 04 03 04
create RSA[]               $09 nc, 08  2A 86 48 F7 0D 01 01 01

variable NumberOf_OIDs
variable NumberOf_KnownOIDs

\ OID values encode the first 2 values in the first byte so the decimal dot form 2.5.4.11 --> 55 04 0B
\ To find the meaning of an OID, decode the OID string and go to http://oid-info.com/
: DecodeOIDfirstByte ( c -- value1 value2 )
   #40 /mod  swap
;
: EncodeOIDfirstByte ( value1 value2 -- c )
   swap #40 * +
;

$100 Buffer: DecodeOIDbuffer[]

: OIDhex-->OIDdot ( $ -- $' ) 
   over c@ DecodeOIDfirstByte swap 1 (u.r) DecodeOIDbuffer[] place 
   s" ." DecodeOIDbuffer[] append
   1 (u.r) DecodeOIDbuffer[] append
   1 /string over + swap ?do 
      s" ." DecodeOIDbuffer[] append
      i c@ 1 (u.r) DecodeOIDbuffer[] append
   loop
;

: ShowOIDdotForm ( a n -- )   over c@  DecodeOIDfirstByte  swap 1 .dml ." ."  1 .dml  1 /string over + swap ?do  ." ." i c@  1 .dml   loop ;

: ttsdf1  ecdsa-with-SHA256[] ShowOIDdotForm ;

\ Information from http://oid-info.com/      | # | hex bytes   | dot form |
create id-at-organizationUnitName(OU)  $04 nc, 03  55 04 0B    \ 2.5.4.11 organizationUnitName  Annex A of Rec. ITU-T X.520 (February 2001) and in ISO/IEC 9594-6: 2001: "The Directory: Selected attribute types". 
create id-at-commonName(CN)            $04 nc, 03  55 04 03    \ 2.5.4.3  Rec. ITU-T X.520 | ISO/IEC 9594-6 "The Directory: Selected attribute types". 
create uniqueIdentifier                $04 nc, 03  55 04 2D    \ 2.5.4.45 uniqueIdentifier  Annex A of Rec. ITU-T X.520 (February 2001) and in ISO/IEC 9594-6: 2001: "The Directory: Selected attribute types

create X509v3_SubjectKeyIdentifier     $04 nc, 03  55 1D 0E    \ 2.5.29.14 subjectKeyIdentifier SubjectKeyIdentifier
create X509v3_CertificatePolicies      $04 nc, 03  55 1D 20    \ 2.5.29.32 certificatePolicies X509v3 Certificate Policies   : critical  Policy: 1.2.826.0.1.1800000.1.2.1.2
create X509v3_AuthorityKeyIdentifier   $04 nc, 03  55 1D 23    \ 2.5.29.35 authorityKeyIdentifier X509v3 Authority Key Identifier
create X509v3_KeyUsage                 $04 nc, 03  55 1D 0F    \ 2.5.29.15 keyUsage X509v3 Key Usage              : critical  Certificate Sign, CRL Sign
create X509v3_SubjectAlternativeName   $04 nc, 03  55 1D 11    \ 2.5.29.17 subjectAltName X509v3 Subject Alternative Name
create X509v3_BasicConstraints         $04 nc, 03  55 1D 13    \ 2.5.29.19 basicConstraints X509v3 Basic Constraints      : critical  CA:TRUE, pathlen:0
create id-ecPublicKey                  $08 nc, 07  2A 86 48 CE 3D 02 01     \ 1.2.134.72.206.61.2.1   id-ecPublicKey
create secp256r1/prime256v1(7)         $09 nc, 08  2A 86 48 CE 3D 03 01 07  \ 1.2.134.72.206.61.3.1.7 secp256r1/prime256v1(7)
create secp256r1/prime256v1(7)Broken   $09 nc, 08  2B 86 48 CE 3D 03 01 07  \ 1.3.134.72.206.61.3.1.7 secp256r1/prime256v1(7) broken by changing the first octet
create ecdsa-with-SHA256               $09 nc, 08  2A 86 48 CE 3D 04 03 02  \ 1.2.134.72.206.61.4.3.2 ecdsa-with-SHA256
create NegTokenInit                    $09 nc, 08  2B 06 01 05 05 07 01 14  \ 1.3.6.1.5.5.7.1.20      NegTokenInit   

: ttsdot ( -- )   
   cr ." id-at-organizationUnitName(OU)"  id-at-organizationUnitName(OU) count 2dup 2 spaces ShowOIDdotForm  56 ToColumn dumpL 
   cr ." id-at-commonName(CN)          "  id-at-commonName(CN)           count 2dup 2 spaces ShowOIDdotForm  56 ToColumn dumpL 
   cr ." uniqueIdentifier              "  uniqueIdentifier               count 2dup 2 spaces ShowOIDdotForm  56 ToColumn dumpL 
                                                                                                                           
   cr ." X509v3_SubjectKeyIdentifier   "  X509v3_SubjectKeyIdentifier    count 2dup 2 spaces ShowOIDdotForm  56 ToColumn dumpL 
   cr ." X509v3_CertificatePolicies    "  X509v3_CertificatePolicies     count 2dup 2 spaces ShowOIDdotForm  56 ToColumn dumpL 
   cr ." X509v3_AuthorityKeyIdentifier "  X509v3_AuthorityKeyIdentifier  count 2dup 2 spaces ShowOIDdotForm  56 ToColumn dumpL 
   cr ." X509v3_KeyUsage               "  X509v3_KeyUsage                count 2dup 2 spaces ShowOIDdotForm  56 ToColumn dumpL 
   cr ." X509v3_SubjectAlternativeName "  X509v3_SubjectAlternativeName  count 2dup 2 spaces ShowOIDdotForm  56 ToColumn dumpL 
   cr ." X509v3_BasicConstraints       "  X509v3_BasicConstraints        count 2dup 2 spaces ShowOIDdotForm  56 ToColumn dumpL 
   cr ." id-ecPublicKey                "  id-ecPublicKey                 count 2dup 2 spaces ShowOIDdotForm  56 ToColumn dumpL 
   cr ." secp256r1/prime256v1(7)       "  secp256r1/prime256v1(7)        count 2dup 2 spaces ShowOIDdotForm  56 ToColumn dumpL 
   cr ." secp256r1/prime256v1(7)Broken "  secp256r1/prime256v1(7)Broken  count 2dup 2 spaces ShowOIDdotForm  56 ToColumn dumpL 
   cr ." ecdsa-with-SHA256             "  ecdsa-with-SHA256              count 2dup 2 spaces ShowOIDdotForm  56 ToColumn dumpL 
   cr ." NegTokenInit                  "  NegTokenInit                   count 2dup 2 spaces ShowOIDdotForm  56 ToColumn dumpL 
;


variable NumberOfid-ecPublicKey
variable NumberOfsecp256r1
variable NumberOfecdsa-with-SHA256

\ $100 Buffer: PrefetchValue[] 

: GetTagAndLengthPlus ( -- )

   ReadPayloadDataOffset @ StartOfSequence ! 

   PayloadDataGet8 dup CurrentTag !  $20 and ConstructedPacket ! 
   ASN.1GetLength CurrentTagLength !
   
   CurrentTag @ $3F and TagBin+  \ Note : ignore the Tag Class

   CurrentTag @ $3F and ASN1_OID = if  
      ReadPayloadDataOffset @ >PayloadData CurrentTagLength @  id-ecPublicKey          count compare 0= if  1 NumberOfid-ecPublicKey +!  then
      ReadPayloadDataOffset @ >PayloadData CurrentTagLength @  secp256r1/prime256v1(7) count compare 0= if  1 NumberOfsecp256r1 +!  then
      ReadPayloadDataOffset @ >PayloadData CurrentTagLength @  ecdsa-with-SHA256       count compare 0= if  1 NumberOfecdsa-with-SHA256 +!  then
   then

   begin  ReadPayloadDataOffset @ SequenceEndTOS >= while  SequenceEndPop drop  -1 Indent +!  repeat

   ConstructedPacket @ if 
      ReadPayloadDataOffset @ CurrentTagLength @ +  SequenceEndPush \ EndOfSequence !       
      1 Indent +!  
   then
;

#70 constant LENGTH_TAB

: ShowTagValue ( -- )
   ReadPayloadDataOffset @ >r 
   LENGTH_TAB #02 + ToColumn  ." | " 
   CurrentTagLength @ 0 ?do  
      i $1F and 0= i 0= not and if  cr LENGTH_TAB #04 + ToColumn  ." | "  then  
      PayloadDataGet8 2.hex  space  
      \ PayloadDataGet8 emit
   loop 

   CurrentTag @ $3F and ASN1_OID = if  
      1 NumberOf_OIDs +!
      r@ >PayloadData CurrentTagLength @  id-ecPublicKey          count compare 0= if  MAGENTA.  ." id-ecPublicKey "          1 NumberOf_KnownOIDs +! RESTORE.  then
      r@ >PayloadData CurrentTagLength @  secp256r1/prime256v1(7) count compare 0= if  MAGENTA.  ." secp256r1/prime256v1(7) " 1 NumberOf_KnownOIDs +! RESTORE.  then
      r@ >PayloadData CurrentTagLength @  ecdsa-with-SHA256       count compare 0= if  MAGENTA.  ." ecdsa-with-SHA256 "       1 NumberOf_KnownOIDs +! RESTORE.  then
      r@ >PayloadData CurrentTagLength @  NegTokenInit            count compare 0= if  MAGENTA.  ." NegTokenInit "            1 NumberOf_KnownOIDs +! RESTORE.  then

      r@ >PayloadData CurrentTagLength @  id-at-organizationUnitName(OU) count compare 0= if  MAGENTA.  ." id-at-organizationUnitName(OU) "  1 NumberOf_KnownOIDs +! RESTORE.  then
      r@ >PayloadData CurrentTagLength @  id-at-commonName(CN)           count compare 0= if  MAGENTA.  ." id-at-commonName(CN)"             1 NumberOf_KnownOIDs +! RESTORE.  then
      r@ >PayloadData CurrentTagLength @  uniqueIdentifier               count compare 0= if  MAGENTA.  ." uniqueIdentifier "                1 NumberOf_KnownOIDs +! RESTORE.  then
      r@ >PayloadData CurrentTagLength @  X509v3_SubjectKeyIdentifier    count compare 0= if  MAGENTA.  ." X509v3_SubjectKeyIdentifier "     1 NumberOf_KnownOIDs +! RESTORE.  then
      r@ >PayloadData CurrentTagLength @  X509v3_CertificatePolicies     count compare 0= if  MAGENTA.  ." X509v3_CertificatePolicies "      1 NumberOf_KnownOIDs +! RESTORE.  then
      r@ >PayloadData CurrentTagLength @  X509v3_AuthorityKeyIdentifier  count compare 0= if  MAGENTA.  ." X509v3_AuthorityKeyIdentifier "   1 NumberOf_KnownOIDs +! RESTORE.  then
      r@ >PayloadData CurrentTagLength @  X509v3_KeyUsage                count compare 0= if  MAGENTA.  ." X509v3_KeyUsage "                 1 NumberOf_KnownOIDs +! RESTORE.  then
      r@ >PayloadData CurrentTagLength @  X509v3_SubjectAlternativeName  count compare 0= if  MAGENTA.  ." X509v3_SubjectAlternativeName "   1 NumberOf_KnownOIDs +! RESTORE.  then
      r@ >PayloadData CurrentTagLength @  X509v3_BasicConstraints        count compare 0= if  MAGENTA.  ." X509v3_BasicConstraints "         1 NumberOf_KnownOIDs +! RESTORE.  then
   then

   r> drop
;

: ShowOneTLV ( -- ) 
   ConstructedPacket @ if 
      Indent @ >r  -1 Indent +!  cr+  r> Indent !
   else
      cr+
   then
   GREEN.  StartOfSequence @ 4.hex  2 spaces  RESTORE. 
   CurrentTag @ ShowTag  LENGTH_TAB ToColumn BLUE. CurrentTagLength @ 3 .hex  RESTORE.  
   \ ." End = "  BLUE. SequenceEndTOS 3 .hex  RESTORE.  2 spaces  ." StackPtr = "  MAGENTA. SequenceEndStackPtr @ 1 .hex  RESTORE. 2 spaces 
   ConstructedPacket @ 0= if 
      ShowTagValue
   then
;

variable CertificateSize

variable BitStringLength$42
variable BitStringLength$45
variable BitStringLength$46
variable BitStringLength$47
variable BitStringLength$48
variable BitStringLength$49
variable BitStringLengthTotal

variable TotalBitStringLength$42
variable TotalBitStringLength$45
variable TotalBitStringLength$46
variable TotalBitStringLength$47
variable TotalBitStringLength$48
variable TotalBitStringLength$49
variable TotalBitStringLengthTotal

: ShowTLVs ( -- ) 
   InitSequenceEndStack 
   TagBinsInit
   0 NumberOfid-ecPublicKey !
   0 NumberOfsecp256r1 !
   0 NumberOfecdsa-with-SHA256 !
   0 NumberOf_OIDs !
   0 NumberOf_KnownOIDs !
   0 BitStringLength$42 !
   0 BitStringLength$45 !  
   0 BitStringLength$46 !  
   0 BitStringLength$47 !  
   0 BitStringLength$48 !  
   0 BitStringLength$49 !  
   0 BitStringLengthTotal !
   -1 Indent ! 
   begin
      GetTagAndLengthPlus ShowOneTLV  
   ReadPayloadDataOffset @ CertificateSize @ >= until
;

\ *****************************************************************************
\ *****************************************************************************

: DumpRelative ( a n -- )
   0 ?do  i $0F and 0= if  cr  i 3 .hex  ."  | "  then  dup i + c@  2.hex  space  loop  drop
;

variable NumberFilesThatFailed 

: LoadPayloadData ( a n -- )
   0 Indent ! 
   PayloadDataInit
   dup CertificateSize ! 
   ( a n -- ) over + swap ?do  i c@ PayloadDataPut8  loop
;

: LoadPayloadFile ( $filename -- ) 
   cr 2dup MAGENTA. type  RESTORE.
\   MyFileBuffer[] MyFileSize @ DumpRelative
   ReadMyFile
   MyFileBuffer[] MyFileSize @ LoadPayloadData
;

include ParseCertificate.f

|PayloadData| Qbuffer: SavedPayloadData[]
Qvariable SavedWritePayloadDataOffset
Qvariable SavedReadPayloadDataOffset

: ShowMyTLVs ( a n -- )
   PayloadData[] SavedPayloadData[] |PayloadData| move
   WritePayloadDataOffset @ SavedWritePayloadDataOffset !
   ReadPayloadDataOffset @  SavedReadPayloadDataOffset !

   LoadPayloadData
   ShowTLVs

   SavedPayloadData[] PayloadData[] |PayloadData| move
   SavedWritePayloadDataOffset @ WritePayloadDataOffset !
   SavedReadPayloadDataOffset @  ReadPayloadDataOffset !
;

: ShowHexASCII ( a n -- )
   over + swap ?do  i c@ 2.hex loop
;

: ShowHexASCIIstruct ( a n -- )
   $05 2.hex $00 2.hex                                \ fixed key ID and location in Key storage as 16 bit Littel Endian value
   MyFileSize @ dup $FF and 2.hex >< $FF and 2.hex    \ the size of the Certificate as 16 bit Littel Endian value
   over + swap ?do  i c@ 2.hex loop
;


: ShowCstyle ( a n -- )
   cr ." const U8 Certificate_"  MyFileNameSaved[] count -EXT -PATH type 
   0 max  $200 min  2dup dup ." [ 0x"  4.hex ."  ] = { "  \ } 
      ( n -- ) 1- 0 max 0 ?do  i $0F and 0= if  cr  4 spaces  then dup i + c@  ." 0x" 2.hex ." u , " 
   loop 
   + 1- 0 max  c@  ." 0x" 2.hex ." u "
   cr ." }; "
;

MAX_PATH Buffer: SaveFilename2

: ShowCert ( $filename -- ) 
   LoadPayloadFile
   ['] ShowTLVs catch if  -1 NumberFilesThatFailed  then 

   cr ." NumberOf_OIDs = " NumberOf_OIDs @ dup .   ."   NumberOf_KnownOIDs = "  NumberOf_KnownOIDs @ dup .  
      = not if  cr  RED ." Not the right number of NumberOf_KnownOIDs !!!"  RESTORE.  ekey #27 = if  quit  then  then

   cr ." id-ecPublicKey    # = "  NumberOfid-ecPublicKey @ BLUE. dup . RESTORE.
      1 = not if  cr  RED ." Not the right number of id-ecPublicKey's in the bins!!!"  RESTORE.  ekey #27 = if  quit  then  then
   cr ." secp256r1/prime256v1(7)         # = "  NumberOfsecp256r1 @ BLUE. dup . RESTORE.
      1 = not if  cr  RED ." Not the right number of secp256r1's in the bins!!!"  RESTORE.  ekey #27 = if  quit  then  then
   cr ." ecdsa-with-SHA256 # = "  NumberOfecdsa-with-SHA256 @ BLUE. dup . RESTORE.
      2 = not if  cr  RED ." Not the right number of ecdsa-with-SHA256's in the bins!!!"  RESTORE.  ekey #27 = if  quit  then  then

   ASN1_UTC_TIME TagBin@ ASN1_GENERALIZED_TIME TagBin@ + 2 = not if cr  RED ." Not the right number of Primitive Times's in the bins!!!"  RESTORE.  ekey #27 = if  quit  then  then
   ASN1_UTC_TIME $20 or TagBin@ ASN1_GENERALIZED_TIME $20 or TagBin@ + 0 = not if cr  RED ." Not the right number of Constructed Times's in the bins!!!"  RESTORE.  ekey #27 = if  quit  then  then

{ 
   cr ." Constructed ASN1_SEQUENCE # = " ASN1_SEQUENCE  $20 or TagBin@ BLUE. dup . RESTORE.
   #16 = NOT if  cr  RED ." Not the right number of Constructed ASN1_SEQUENCE's in the bins!!!"  RESTORE.  ekey #27 = if  quit  then  then 
   cr ." Primitive ASN1_BIT_STRING # = " ASN1_BIT_STRING TagBin@ BLUE. dup . RESTORE.
   ."    Constructed ASN1_BIT_STRING # = " ASN1_BIT_STRING $20 or TagBin@ BLUE. dup . RESTORE. 
   1 = swap 2 = and not if  cr  RED ." Not the right number in the bins!!!"  RESTORE.  ekey #27 = if  quit  then  then
}
\   cr ." Done......." 
   cr
\   ShowSequenceStack
   PayloadDataInit
   MyFileBuffer[] MyFileSize @ Certificate_getPublicKey swap drop $40 = not if  cr RED. ." Certificate_getPublicKey not $40 bytes!!!"    RESTORE.  quit  then 
   MyFileBuffer[] MyFileSize @ Certificate_getAuthorityKeyIdentifier swap drop $08 = not if  cr RED. ." Certificate_getAuthorityKeyIdentifier not $08 bytes!!!"  RESTORE.  quit  then 

   SaveFilename count SaveFilename2 place

\ save a HexASCII format file with ".txt" appended to the filename
   InitLogBuffer
   MyFileNameSaved[] count SaveFilename place
   s" .txt" SaveFilename append
   +Log
   MyFileBuffer[] MyFileSize @ ShowHexASCII
   -Log
   LogToFile
   cr ." *.txt File saved = " BLUE. SaveFilename count type  RESTORE.
   SaveFilename2 count SaveFilename place

   \ save a HexASCII format file with ".txt" appended to the filename
   InitLogBuffer
   MyFileNameSaved[] count SaveFilename place
   s" .Struct.txt" SaveFilename append
   +Log
   MyFileBuffer[] MyFileSize @ ShowHexASCIIstruct
   -Log
   LogToFile
   cr ." *.txt File saved = " BLUE. SaveFilename count type  RESTORE.
   SaveFilename2 count SaveFilename place

   \ save a HexASCII format file with ".txt" appended to the filename
   InitLogBuffer
   MyFileNameSaved[] count SaveFilename place
   s" .c" SaveFilename append
   +Log
   MyFileBuffer[] MyFileSize @ ShowCstyle
   -Log
   LogToFile
   cr ." *.c File saved = " BLUE. SaveFilename count type  RESTORE.
   SaveFilename2 count SaveFilename place
;

: ttsc1 ( -- )   s" ICA_OrgCA.der" ShowCert ;
create sc1 $10 nc,  30 82 02 2D 30 82 01 D3 A0 03 02 01 02 02 10 10

: ttsc2 ( -- )   s" org-root.der" ShowCert ;
create sc2 $10 nc,  30 82 01 9E 30 82 01 44 A0 03 02 01 02 02 10 10


MAX_PATH 2* Buffer: MyLongDirectoryBuffer
MAX_PATH 2* Buffer: MyLongDirectoryBufferSaved

\ perform the action of xt for each file in the current folder, recursively
: WalkFiles2 ( xt -- )  
   is FileAction  
   
   MAX_PATH MyLongDirectoryBufferSaved GetCurrentDirectory drop   \ save our current directory
   MyLongDirectoryBuffer SetCurrentDirectory drop

   0 ShowZeroFileLengthError !
   depth lastStackDepth !

   s" *.DER" FileSpecification zplace 
   Walk drop
   depth lastStackDepth @ - StackDepthValue !
   -1 ShowZeroFileLengthError ! 
   
   MyLongDirectoryBufferSaved SetCurrentDirectory drop
; 

: ShowDirFileName2 ( -- ) 
   MyLongPathBuffer zcount  dup >r type  60 r> - 0 max spaces  
;

: .Filename2 ( zstr -- )
   SetDirFileName
   ShowDirFileName
;

variable NumberOfFiles

: MyAction ( zstr -- f )
   SetDirFileName
\   cr MyLongPathBuffer zcount type 
   MyLongPathBuffer zcount ShowCert

   1 NumberOfFiles +!
   0
;

: ShowTotalBitStringFieldSizes ( -- )
   cr ." TotalBitStringLength$42 = " 
                      TotalBitStringLength$42 @ dup .  
      ."   ...$45 = " TotalBitStringLength$45 @ dup . + 
      ."   ...$46 = " TotalBitStringLength$46 @ dup . + 
      ."   ...$47 = " TotalBitStringLength$47 @ dup . + 
      ."   ...$48 = " TotalBitStringLength$48 @ dup . +  
      ."   ...$49 = " TotalBitStringLength$49 @ dup . + 
      ."  Others = "  TotalBitStringLengthTotal @ swap - dup . 
   0= not if RED. cr ." TotalBitStringLength other != 0 !!!"  RESTORE.  quit  then
;


: ttAll ( -- )

   0 NumberOfFiles !  0 NumberFilesThatFailed !
   0 TotalBitStringLength$42 !
   0 TotalBitStringLength$45 !
   0 TotalBitStringLength$46 !
   0 TotalBitStringLength$47 !
   0 TotalBitStringLength$48 !
   0 TotalBitStringLength$49 !
   0 TotalBitStringLengthTotal !

   s" C:\Users\oakfordh\Documents\Certificates\smki_cert_run_130115" MyLongDirectoryBuffer zplace
   s" *.c*" FileSpecification zplace   \ default for WalkFiles2 
   ['] MyAction WalkFiles2
   
   cr ." NumberOfFiles = " NumberOfFiles @ .  2 spaces  ." NumberFilesThatFailed = " NumberFilesThatFailed @ .
   cr 
   cr ShowTotalBitStringFieldSizes
;

\ 009A  03 Universal        P BIT STRING                                        
create PublicKey[]1  
   $43 nc, 42  00 04 FF 8D B1 EC FB 5E 7D B3 70 22 58 32 1D 28 F8 3B D4 91 00 91 5F 08 43 81 47 91 6F 7F 69 6A 87 C7 57 66 6E D4 59 16 6E 81 19 5A 3A 68 6E 9F 4E D3 C5 56 B1 24 F8 26 28 D9 3E 2E 57 7D C0 2F 07 E1   

create PublicKey[]2  
\ 0077  03 Universal        P BIT STRING                     
   $43 nc, 42  00 04 08 71 8F 50 D8 9D 39 10 C2 BD 54 27 44 B0 AA 22 93 0E C8 9C 8E 97 79 FE CD 7E 57 96 A7 71 27 E9 81 C4 F7 CB FC 8B D6 0C 6C CA C0 68 5B B1 C9 91 95 80 2B 60 91 63 79 C4 D8 2C F6 DA 8B 65 EA 98

create PublicKey[]3  
\ 0077  03 Universal        P BIT STRING                     
   $43 nc, 42  00 04 28 08 98 9B AB 0C 4D 39 98 B1 B8 28 83 95 E4 7E 56 5D C2 6D DC 90 D9 C6 64 43 85 17 1B 74 DF D3 2F CC 38 0D CC A7 6E 44 97 A3 60 FF F0 76 6B 5F 84 A0 4A B8 49 5B 42 AD DB 9B 45 6B 15 D9 70 6C

\ 014E  06 Universal        P OBJECT IDENTIFIER                   8 | 2A 86 48 CE 3D 04 03 02 ecdsa-with-SHA256
\    0158  03 Universal        P BIT STRING                             
create Signature[]3          
   $48 nc, 47  00 30 44 02 20 26 94 B0 84 1D D9 F3 45 3B 52 E2 38 81 0D 49 AA A0 38 F4 DF F1 CE 59 06 10 E1 FA E6 8A DF BD 79 02 20 33 82 13 E9 A1 B5 22 A1 03 E9 75 72 A9 F3 61 4E 35 DB 57 AA 9B 4D AC 95 E9 5E C6 F0 C2 99 63 21    

\  0158  03 Universal        P BIT STRING                             
create Signature[]1   $49 nc, 48  00 30 45 02 20 77 EA 09 85 83 89 C9 6F 0A C4 F5 05 72 A6 AD 78 4B D4 EE 8A 02 90 EF 26 EA 0C 13 B8 FE A9 5A F7 02 21 00 A5 D1 2E 11 CD 47 91 40 00 4D 2E 77 57 E7 59 AC 40 70 A6 AF 09 2A 93 D1 91 A7 53 6C ED AF 35 4E


\ 00E4  06 Universal        P OBJECT IDENTIFIER       3 | 55 1D 0E X509v3_SubjectKeyIdentifier
\  00E9  04 Universal        P OCTET STRING            A | 04 08 49 6B 2D 04 D2 90 4E D2
create X509v3_SubjectKeyIdentifier[]1 $0B nc, 0A  04 08 49 6B 2D 04 D2 90 4E D2

create KeyPolicyBit[]1  $05 nc, 04  03 02 01 06
create X509v3_BasicConstraintsBit[]1  $09 nc, 08  30 06 01 01 FF 02 01 00

create X509v3_CertificatePoliciesBit[]4 $13 nc, 12  30 10 30 0E 06 0C 2A 86 3A 00 01 ED EE 40 01 02 01 03
create X509v3_CertificatePoliciesBitOfBit[]4               $0D nc, 0C  2A 86 3A 00 01 ED EE 40 01 02 01 03

: ttcps ( -- )   X509v3_CertificatePoliciesBitOfBit[]4 count 2dup cr dumpL cr over + swap ?do  i c@ 3 .dml space  loop ;


create MESSAGE_1_CORRECT_SIG[]      c," "85AE39D45D5C73A44070DF71C7A0976BAF60A3626E6D08D167AA7CF4AB8393B0B413E91D3E79FD6CCC93F45DB0A20BE5264B5CE9BA56A2470072784DD1A11752";

: ShowSigs ( -- )
   cr ." PublicKey[]1 ICA_OrgCA.der "  
   PublicKey[]1 count  1 /string 1 /string cr dumpL    \ BitString has intial 0 byte which must be removed
   cr
   cr ." PublicKey[]3 device-ee-a0cd12f1a0000001-ds.der " 
   PublicKey[]3 count  1 /string 1 /string cr dumpL    \ BitString has intial 0 byte which must be removed
   cr
   cr ." Signature[]1 ICA_OrgCA.der "  
   Signature[]1 count 1 /string ShowMyTLVs   \ BitString has intial 0 byte which must be removed
   cr
   cr ." Signature[]3 device-ee-a0cd12f1a0000001-ds.der "  
   Signature[]3 count 1 /string ShowMyTLVs    \ BitString has intial 0 byte which must be removed
   cr
   cr ." X509v3_SubjectKeyIdentifier[]1 ICA_OrgCA.der "
   X509v3_SubjectKeyIdentifier[]1 count ShowMyTLVs
;

: ASN1DER  cr ." Type  ttsc1 , ttsc2 , ttall or ShowSigs " ;

MAGENTA.  ASN1DER  RESTORE.

: ttall+  page  cr ." >>>>>>>>>>>>>1 "  ekey drop  ['] ttall catch cr ." catch = " . space   ekey $06 = if  quit  then  bye ;

: ASN1DER+ ( -- )   ASN1DER  
   \ ekey $1B = if  cr ."  ok " cr  quit  then 
   quit 
;

starter ASN1DER+  program ASN1DER.exe  sf.ico


\\ Below here is not loaded ...

{ 
\ Returns the size of the binary string, without the trailing 0 
#define OID_SIZE(x) (sizeof(x) - 1)

/**
 * Compares an asn1_buf structure to a reference OID.
 *
 * Only works for 'defined' oid_str values (OID_HMAC_SHA1), you cannot use a
 * 'unsigned char *oid' here!
 *
 * Warning: returns true when the OIDs are equal (unlike memcmp)!
 
#define OID_CMP(oid_str, oid_buf)                                   \
        ( ( OID_SIZE(oid_str) == (oid_buf)->len ) &&                \
          memcmp( (oid_str), (oid_buf)->p, (oid_buf)->len) == 0 )
}

\\

Example:
Example of information in certificate (after decoding, example from SMKI, organisation Certificate):
$ openssl x509 -inform der -in ICA_OrgCA.der -out ICA_OrgCA.pem
$ openssl x509 -in ICA_OrgCA.pem -noout -text
Certificate:
    Data:
        Version: 3 (0x2)
        Serial Number:
            10:00:00:00:00:00:00:01:10:00:00:00:00:00:00:02
    Signature Algorithm: ecdsa-with-SHA256
        Issuer: OU=00, CN=FF
        Validity
            Not Before: Jan  7 09:20:50 2015 GMT
            Not After : Jan  8 09:20:50 2040 GMT
        Subject: OU=07, CN=FF
        Subject Public Key Info:
            Public Key Algorithm: id-ecPublicKey
                Public-Key: (256 bit)
                pub:
                    04:ff:8d:b1:ec:fb:5e:7d:b3:70:22:58:32:1d:28:
                    f8:3b:d4:91:00:91:5f:08:43:81:47:91:6f:7f:69:
                    6a:87:c7:57:66:6e:d4:59:16:6e:81:19:5a:3a:68:
                    6e:9f:4e:d3:c5:56:b1:24:f8:26:28:d9:3e:2e:57:
                    7d:c0:2f:07:e1
                ASN1 OID: prime256v1
        X509v3 extensions:
            X509v3 Subject Key Identifier:
                49:6B:2D:04:D2:90:4E:D2
            X509v3 Authority Key Identifier:
                keyid:4B:62:32:39:E2:3C:D6:C7

            X509v3 Basic Constraints: critical
                CA:TRUE, pathlen:0
            X509v3 Certificate Policies: critical
                Policy: 1.2.826.0.1.1800000.1.2.1.2

            X509v3 Key Usage: critical
                Certificate Sign, CRL Sign
    Signature Algorithm: ecdsa-with-SHA256
         30:45:02:20:77:ea:09:85:83:89:c9:6f:0a:c4:f5:05:72:a6:
         ad:78:4b:d4:ee:8a:02:90:ef:26:ea:0c:13:b8:fe:a9:5a:f7:
         02:21:00:a5:d1:2e:11:cd:47:91:40:00:4d:2e:77:57:e7:59:
         ac:40:70:a6:af:09:2a:93:d1:91:a7:53:6c:ed:af:35:4e

Example of information in certificate (after decoding, example from SMKI, Device Certificate):
$ openssl x509 -in device-ee-a0cd12f2a0000064-ka.pem -noout -text
Certificate:
    Data:
        Version: 3 (0x2)
        Serial Number:
            10:00:00:00:00:00:00:01:10:00:00:00:00:00:00:c8
    Signature Algorithm: ecdsa-with-SHA256
        Issuer: CN=0002
        Validity
            Not Before: Jan  7 09:13:49 2015 GMT
            Not After : Dec 31 23:59:59 9999 GMT
        Subject:
        Subject Public Key Info:
            Public Key Algorithm: id-ecPublicKey
                Public-Key: (256 bit)
                pub:
                    04:7b:ff:ce:e7:cf:14:94:d8:97:03:88:9d:73:a1:
                    8f:5f:0b:69:d5:12:5f:68:ef:e8:ed:c7:d0:2c:5b:
                    cc:0c:66:ea:13:1c:4c:b2:e2:5d:ec:9e:4b:d2:0f:
                    de:77:ad:ee:8d:52:ff:1b:0c:c5:b7:43:ca:b6:64:
                    59:f2:ca:05:61
                ASN1 OID: prime256v1
        X509v3 extensions:
            X509v3 Subject Key Identifier:
                41:5E:26:6C:99:8F:68:B5
            X509v3 Authority Key Identifier:
                keyid:45:82:9B:69:A8:2C:B9:23

            X509v3 Certificate Policies: critical
                Policy: 1.2.826.0.1.1800000.1.2.1.3

            X509v3 Key Usage: critical
                Key Agreement
            X509v3 Subject Alternative Name: critical
                othername:<unsupported>
    Signature Algorithm: ecdsa-with-SHA256
         30:45:02:20:76:3d:4a:da:b5:19:e2:9b:54:b0:4f:7d:3a:d2:
         1f:e8:48:08:61:8a:c3:c4:f5:03:7c:2c:9c:52:f9:76:61:15:
         02:21:00:90:6b:a0:f3:2c:de:4c:66:e8:51:d6:cf:de:3d:59:
         3f:07:f1:26:90:44:bd:bb:72:4d:f1:1c:69:dc:c9:d1:fc
\\ 

\ Cooper, et al.              Standards Track                    [Page 16]
\ RFC 5280            PKIX Certificate and CRL Profile            May 2008

Basic Certificate Fields

   The X.509 v3 certificate basic syntax is as follows.  For signature
   calculation, the data that is to be signed is encoded using the ASN.1
   distinguished encoding rules (DER) [X.690].  
   ASN.1 DER encoding is a tag, length, value encoding system for each element.

   Certificate  ::=  SEQUENCE  {
        tbsCertificate       TBSCertificate,
        signatureAlgorithm   AlgorithmIdentifier,
        signatureValue       BIT STRING  }

   TBSCertificate  ::=  SEQUENCE  {
        version         [0]  EXPLICIT Version DEFAULT v1,
        serialNumber         CertificateSerialNumber,
        signature            AlgorithmIdentifier,
        issuer               Name,
        validity             Validity,
        subject              Name,
        subjectPublicKeyInfo SubjectPublicKeyInfo,
        issuerUniqueID  [1]  IMPLICIT UniqueIdentifier OPTIONAL,
                             -- If present, version MUST be v2 or v3

        subjectUniqueID [2]  IMPLICIT UniqueIdentifier OPTIONAL,
                             -- If present, version MUST be v2 or v3
        extensions      [3]  EXPLICIT Extensions OPTIONAL
                             -- If present, version MUST be v3
        }

   Version  ::=  INTEGER  {  v1(0), v2(1), v3(2)  }

   CertificateSerialNumber  ::=  INTEGER

   Validity ::= SEQUENCE {
        notBefore      Time,
        notAfter       Time }

   Time ::= CHOICE {
        utcTime        UTCTime,
        generalTime    GeneralizedTime }

   UniqueIdentifier  ::=  BIT STRING

   SubjectPublicKeyInfo  ::=  SEQUENCE  {
        algorithm            AlgorithmIdentifier,
        subjectPublicKey     BIT STRING  }

   Extensions  ::=  SEQUENCE SIZE (1..MAX) OF Extension

   Extension  ::=  SEQUENCE  {
        extnID      OBJECT IDENTIFIER,
        critical    BOOLEAN DEFAULT FALSE,
        extnValue   OCTET STRING
                    -- contains the DER encoding of an ASN.1 value
                    -- corresponding to the extension type identified
                    -- by extnID
        }

   The following items describe the X.509 v3 certificate for use in the
   Internet.



{ 
EOC (End-of-Content)	P     0	0
BOOLEAN	            P     1	1
INTEGER	            P     2	2
BIT STRING	         P/C	3	3
OCTET STRING	      P/C	4	4
NULL	               P     5	5
OBJECT IDENTIFIER	   P     6	6
Object Descriptor	   P/C	7	7
EXTERNAL	            C     8	8
REAL (float)	      P     9	9
ENUMERATED	         P     10	A
EMBEDDED PDV	      C     11	B
UTF8String	         P/C	12	C
RELATIVE-OID	      P     13	D
(reserved)	-              14	E
(reserved)	-              15	F
SEQUENCE 	         C     16	10
SET                  C     17	11
NumericString	      P/C	18	12
PrintableString	   P/C	19	13
T61String	         P/C	20	14
VideotexString	      P/C	21	15
IA5String	         P/C	22	16
UTCTime	            P/C	23	17
GeneralizedTime	   P/C	24	18
GraphicString	      P/C	25	19
VisibleString	      P/C	26	1A
GeneralString	      P/C	27	1B
UniversalString	   P/C	28	1C
CHARACTER STRING	   P/C	29	1D
BMPString	         P/C	30	1E
(use long-form)	-	      31	1F
}
