\ ParseCertificate.f  2015 Mar 10 

variable PayloadDataStart     \ points to the start of the Certificate
variable PayloadDataOffset    \ the offset in the Certificate of the current octet being read
variable PayloadDataLength    \ the length of the Certificate being parsed

: PayloadDataGetCh ( -- c )  
   PayloadDataStart @ PayloadDataOffset @ + c@ 
   1 PayloadDataOffset +!
\   dup MAGENTA. 2.hex space RESTORE.
;

\ The special ASN.1 variable-length encoded length
: ASN.1GetLengthLimited ( -- u ) 
   PayloadDataGetCh dup $80 and if     
      case
      $80 of  -1  endof    \ indefinite length, terminated by and EOC with 0 length ( two 0's )
      $81 of  PayloadDataGetCh  endof
      $82 of  PayloadDataGetCh 8 lshift PayloadDataGetCh or  endof
      $83 of  PayloadDataGetCh 8 lshift PayloadDataGetCh or  8 lshift PayloadDataGetCh or  endof
         cr  RED.  ." Illegal AXDR String length value = "  2 .hex   ."   PayloadDataOffset = "  PayloadDataOffset @ 4 .hex  RESTORE. 
         quit
      endcase
   then
; 

variable FoundRequiredOID         \ set true when the reqruied OID is found
variable vRequiredOID

: SkipValue ( -- )
   CurrentTagLength @ PayloadDataOffset +!  \ skip the value field of the current Tag
;

: GetTagAndLength ( -- )
   PayloadDataGetCh CurrentTag !    
   ASN.1GetLengthLimited CurrentTagLength ! 
;

: GetNextTLV ( -- )  
   SkipValue         \ skip the current Tag's value
   GetTagAndLength   \ load the next Tag
   PayloadDataOffset @ CurrentTagPtr ! 
;

: GetTagLengthValue ( -- f )
   
   GetTagAndLength

   CurrentTag @ $20 and if    \ ConstructedPacket, no data, just the length of SEQUENCE or SET
      \ PayloadDataOffset @ CurrentTagLength @ +  SequenceEndPush 
   else  \ primitive packet, contains real data to check and skip 

      CurrentTag @ $3F and ASN1_OID = if 
         vRequiredOID @ count  PayloadDataStart @ PayloadDataOffset @ +  CurrentTagLength @  compare 0= if
            GetNextTLV
            1 FoundRequiredOID !
         then
      then
      
      FoundRequiredOID @ 0= if  
         SkipValue   \ skip the current Tag's value, step to the next TLV
      then
   then
   FoundRequiredOID @
;

variable vSkipOctets 

\ given the address and Length of a Certificate, the required OID, returns the address and length of the required Value
\ SkipOctets is the number of octets that must be skipped to get to the required value.
\ Please see below the words that use Certificate_getOID for more details.
: Certificate_getOID ( $Certificate sRequiredOID SkipOctets -- Address Length | 0 0 )  
   ( SkipOctets -- ) vSkipOctets !
   ( $OIDoffset ) vRequiredOID !
   ( n ) PayloadDataLength ! 
   ( a ) PayloadDataStart ! 
   
   0 PayloadDataOffset !   \ start at the begining of the Certificate
   0 FoundRequiredOID !    \ not found yet

   0 0   \ failed address and count in case we don't find the OID

   begin
      
      GetTagLengthValue if   \ found
         2drop    \ removet the failed address and count 
         PayloadDataStart @ CurrentTagPtr @ + 
         CurrentTagLength @ 
         vSkipOctets @ /string  \ skip the first n octets from the start of the string
      then      
   
   PayloadDataOffset @ PayloadDataLength @ >= FoundRequiredOID @ or until
;

\ Gets the certificate's Public Key
\ There is only one secp256r1/prime256v1(7) OID in a Certificate (which always follows the OID_id_ecPublicKey ) 
\ The required OID's value is the first Tag following secp256r1/prime256v1(7)
\ The PublicKey value follows the initial BitString octet and a value of 0x04, so skip 0x02u octets
: Certificate_getPublicKey ( $Certificate -- $TagValue )
   secp256r1/prime256v1(7) $02 Certificate_getOID  
;

\ Gets the certificate's AuthorityKeyIdentifier ( Originator ID )
\ There is only one OID_X509v3_AuthorityKeyIdentifier in a Certificate 
\ The required OID's value is the first Tag following it
\ X509v3_AuthorityKeyIdentifier := 
\ 0000  30 Universal        C SEQUENCE                                    A
\     0002  80 Context-specific P EOC (End-of-Content)                    8 | 4B 62 32 39 E2 3C D6 C7  ok
\ The  AuthorityKeyIdentifier value follows a Sequence|Length pair and Context-specific P EOC (End-of-Content) Sequence|Length pair , so skip 0x04u octets
\ This assumes that the OID_X509v3_AuthorityKeyIdentifier is less thatn ( 127 - 4 ) octets long, which is always the case.
: Certificate_getAuthorityKeyIdentifier ( $Certificate -- $TagValue )
   X509v3_AuthorityKeyIdentifier $04 Certificate_getOID 
;

: ttPCx ( -- )   s" ICA_OrgCA.der" LoadPayloadFile  MyFileBuffer[] MyFileSize @    secp256r1/prime256v1(7)Broken  $02 Certificate_getOID  dumpL ;

: ttPC1 ( -- )   s" ICA_OrgCA.der" LoadPayloadFile  MyFileBuffer[] MyFileSize @ Certificate_getPublicKey dumpL ;
: ttPC2 ( -- )   s" ICA_OrgCA.der" LoadPayloadFile  MyFileBuffer[] MyFileSize @ Certificate_getAuthorityKeyIdentifier dumpL ;

: ttPC3 ( -- )   s" org-root.der"  LoadPayloadFile  MyFileBuffer[] MyFileSize @ Certificate_getPublicKey dumpL ;
: ttPC4 ( -- )   s" org-root.der"  LoadPayloadFile  MyFileBuffer[] MyFileSize @ Certificate_getAuthorityKeyIdentifier dumpL ;

\ dump on the same line, up to 64 bytes, then move to the next line
: dumpL+++ ( a n -- )   
   0 max  $200 min  space  dup ."  MyArray[ "  3 .hex ."  ] = { "  \ } 
      ( n -- ) 0 ?do  i $0F and 0= if  cr  4 spaces  then dup i + c@  ." 0x" 2.hex ." u , " 
   loop 
   cr ." }; "
; 


\\ 

/* From : https://www.ietf.org/rfc/rfc2459.txt 
Basic Certificate Fields

   The X.509 v3 certificate basic syntax is as follows.  For signature
   calculation, the certificate is encoded using the ASN.1 distinguished
   encoding rules (DER) [X.208].  ASN.1 DER encoding is a tag, length,
   value encoding system for each element.

   Certificate  ::=  SEQUENCE  {
        tbsCertificate       TBSCertificate,
        signatureAlgorithm   AlgorithmIdentifier,
        signatureValue       BIT STRING  }

   TBSCertificate  ::=  SEQUENCE  {
        version         [0]  EXPLICIT Version DEFAULT v1,
        serialNumber         CertificateSerialNumber,
        signature            AlgorithmIdentifier,
        issuer               Name,
        validity             Validity,
        subject              Name,
        subjectPublicKeyInfo SubjectPublicKeyInfo,
        issuerUniqueID  [1]  IMPLICIT UniqueIdentifier OPTIONAL,
                             -- If present, version shall be v2 or v3
        subjectUniqueID [2]  IMPLICIT UniqueIdentifier OPTIONAL,
                             -- If present, version shall be v2 or v3
        extensions      [3]  EXPLICIT Extensions OPTIONAL
                             -- If present, version shall be v3
        }

   Version  ::=  INTEGER  {  v1(0), v2(1), v3(2)  }

   CertificateSerialNumber  ::=  INTEGER

   Validity ::= SEQUENCE {
        notBefore      Time,
        notAfter       Time }

   Time ::= CHOICE {
        utcTime        UTCTime,
        generalTime    GeneralizedTime }

   UniqueIdentifier  ::=  BIT STRING

   SubjectPublicKeyInfo  ::=  SEQUENCE  {
        algorithm            AlgorithmIdentifier,
        subjectPublicKey     BIT STRING  }

   Extensions  ::=  SEQUENCE SIZE (1..MAX) OF Extension

   Extension  ::=  SEQUENCE  {
        extnID      OBJECT IDENTIFIER,
        critical    BOOLEAN DEFAULT FALSE,
        extnValue   OCTET STRING  }

   The following items describe the X.509 v3 certificate for use in the
   Internet.

4.1.1  Certificate Fields

   The Certificate is a SEQUENCE of three required fields. The fields
   are described in detail in the following subsections.

4.1.1.1  tbsCertificate

   The field contains the names of the subject and issuer, a public key
   associated with the subject, a validity period, and other associated
   information.  The fields are described in detail in section 4.1.2;
   the tbscertificate may also include extensions which are described in
   section 4.2.

4.1.1.2  signatureAlgorithm

   The signatureAlgorithm field contains the identifier for the
   cryptographic algorithm used by the CA to sign this certificate.
   Section 7.2 lists the supported signature algorithms.

   An algorithm identifier is defined by the following ASN.1 structure:

   AlgorithmIdentifier  ::=  SEQUENCE  {
        algorithm               OBJECT IDENTIFIER,
        parameters              ANY DEFINED BY algorithm OPTIONAL  }

   The algorithm identifier is used to identify a cryptographic
   algorithm.  The OBJECT IDENTIFIER component identifies the algorithm
   (such as DSA with SHA-1).  The contents of the optional parameters
   field will vary according to the algorithm identified. Section 7.2
   lists the supported algorithms for this specification.

   This field MUST contain the same algorithm identifier as the
   signature field in the sequence tbsCertificate (see sec. 4.1.2.3).

4.1.1.3  signatureValue

   The signatureValue field contains a digital signature computed upon
   the ASN.1 DER encoded tbsCertificate.  The ASN.1 DER encoded
   tbsCertificate is used as the input to the signature function. This
   signature value is then ASN.1 encoded as a BIT STRING and included in
   the Certificate's signature field. The details of this process are
   specified for each of the supported algorithms in Section 7.2.

   By generating this signature, a CA certifies the validity of the
   information in the tbsCertificate field.  In particular, the CA
   certifies the binding between the public key material and the subject
   of the certificate.

4.1.2  TBSCertificate

   The sequence TBSCertificate contains information associated with the
   subject of the certificate and the CA who issued it.  Every
   TBSCertificate contains the names of the subject and issuer, a public
   key associated with the subject, a validity period, a version number,
   and a serial number; some may contain optional unique identifier
   fields.  The remainder of this section describes the syntax and
   semantics of these fields.  A TBSCertificate may also include
   extensions.  Extensions for the Internet PKI are described in Section
   4.2.

4.1.2.1  Version

   This field describes the version of the encoded certificate.  When
   extensions are used, as expected in this profile, use X.509 version 3
   (value is 2).  If no extensions are present, but a UniqueIdentifier
   is present, use version 2 (value is 1).  If only basic fields are
   present, use version 1 (the value is omitted from the certificate as
   the default value).

   Implementations SHOULD be prepared to accept any version certificate.
   At a minimum, conforming implementations MUST recognize version 3
   certificates.

   Generation of version 2 certificates is not expected by
   implementations based on this profile.

4.1.2.2  Serial number

   The serial number is an integer assigned by the CA to each
   certificate.  It MUST be unique for each certificate issued by a
   given CA (i.e., the issuer name and serial number identify a unique
   certificate).

4.1.2.3  Signature

   This field contains the algorithm identifier for the algorithm used
   by the CA to sign the certificate.

   This field MUST contain the same algorithm identifier as the
   signatureAlgorithm field in the sequence Certificate (see sec.
   4.1.1.2).  The contents of the optional parameters field will vary
   according to the algorithm identified.  Section 7.2 lists the
   supported signature algorithms.

4.1.2.4  Issuer

   The issuer field identifies the entity who has signed and issued the
   certificate.  The issuer field MUST contain a non-empty distinguished
   name (DN).  The issuer field is defined as the X.501 type Name.
   [X.501] Name is defined by the following ASN.1 structures:

   Name ::= CHOICE {
     RDNSequence }

   RDNSequence ::= SEQUENCE OF RelativeDistinguishedName

   RelativeDistinguishedName ::=
     SET OF AttributeTypeAndValue

   AttributeTypeAndValue ::= SEQUENCE {
     type     AttributeType,
     value    AttributeValue }

   AttributeType ::= OBJECT IDENTIFIER

   AttributeValue ::= ANY DEFINED BY AttributeType

   DirectoryString ::= CHOICE {
         teletexString           TeletexString (SIZE (1..MAX)),
         printableString         PrintableString (SIZE (1..MAX)),
         universalString         UniversalString (SIZE (1..MAX)),
         utf8String              UTF8String (SIZE (1.. MAX)),
         bmpString               BMPString (SIZE (1..MAX)) }

   The Name describes a hierarchical name composed of attributes, such
   as country name, and corresponding values, such as US.  The type of
   the component AttributeValue is determined by the AttributeType; in
   general it will be a DirectoryString.

   The DirectoryString type is defined as a choice of PrintableString,
   TeletexString, BMPString, UTF8String, and UniversalString.  The
   UTF8String encoding is the preferred encoding, and all certificates
   issued after December 31, 2003 MUST use the UTF8String encoding of
   DirectoryString (except as noted below).  Until that date, conforming
   CAs MUST choose from the following options when creating a
   distinguished name, including their own:

      (a) if the character set is sufficient, the string MAY be
      represented as a PrintableString;

      (b) failing (a), if the BMPString character set is sufficient the
      string MAY be represented as a BMPString; and

      (c) failing (a) and (b), the string MUST be represented as a
      UTF8String.  If (a) or (b) is satisfied, the CA MAY still choose
      to represent the string as a UTF8String.
*/


/* Unfinished, because they are almost certainly not required...

// Based on the Java API from U16 https://docs.oracle.com/javame/8.0/api/gcf/api/javax/microedition/pki/Certificate.html
getIssuer()        // Gets the name of this certificate's issuer.
getNotAfter()      // Gets the time after which this Certificate may not be used from the validity period.
getNotBefore()     // Gets the time before which this Certificate may not be used from the validity period.
getSerialNumber()  // Gets the printable form of the serial number of this Certificate.
getSigAlgName()    // Gets the name of the algorithm used to sign the Certificate.
getSubject()       // Gets the name of this certificate's subject.
getType()          // Get the type of the Certificate.
getVersion()       // Gets the version number of this Certificate.


*/